try:
    from fastai.vision.data import imagenet_stats
    from fastai.vision.transform import crop, dihedral_affine
    import torch
    from pathlib import Path
    from functools import partial
    import xml.etree.ElementTree as ET
    from .models._ssd_utils import SSDObjectItemList
    HAS_FASTAI = True
except:
    HAS_FASTAI = False

def _raise_fastai_import_error():
    raise Exception('This class requires fastai, PyTorch and torchvision as its dependencies. Install it using "conda install -c pytorch -c fastai fastai pytorch torchvision"')

def _bb_pad_collate(samples, pad_idx=0):
    "Function that collect `samples` of labelled bboxes and adds padding with `pad_idx`."
    arr = []
    for s in samples:
        try:
            arr.append(len(s[1].data[1]))
        except Exception as e:
            # set_trace()
            # print(s[1].data[1],s[1].data[1],e)
            arr.append(0)
    max_len = max(arr)
#    max_len = max([len(s[1].data[1]) for s in samples])
    bboxes = torch.zeros(len(samples), max_len, 4)
    labels = torch.zeros(len(samples), max_len).long() + pad_idx
    imgs = []
    for i,s in enumerate(samples):
        imgs.append(s[0].data[None])
        bbs, lbls = s[1].data
        # print(bbs, lbls)
        try:
            bboxes[i,-len(lbls):] = bbs
            labels[i,-len(lbls):] = lbls
        except Exception as e:
            pass
    return torch.cat(imgs,0), (bboxes,labels)


def _get_bbox_lbls(imagefile, class_mapping):
    xmlfile = imagefile.parents[1] / 'labels' / imagefile.name.replace('{ims}'.format(ims=imagefile.suffix), '.xml')
    tree = ET.parse(xmlfile)
    xmlroot = tree.getroot()
    bboxes  = []
    classes = []
    for child in xmlroot:
        if child.tag == 'object':
            xmin, ymin, xmax, ymax = float(child[1][0].text),\
            float(child[1][1].text),\
            float(child[1][2].text),\
            float(child[1][3].text)
            bboxes.append([ymin, xmin, ymax, xmax])
            classes.append(class_mapping[int(child[0].text)])

    return [bboxes, classes]

def prepare_data(path, class_mapping, chip_size=224, val_split_pct=0.1, batch_size=64, transforms=None, collate_fn=_bb_pad_collate, seed=42):
    """
    Prepares a Fast.ai DataBunch from the exported Pascal VOC image chips
    exported by Export Training Data tool in ArcGIS Pro or Image Server.
    This DataBunch consists of training and validation DataLoaders with the
    specified transformations, chip size, batch size, split percentage.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    path                    Required string. Path to data directory.
    ---------------------   -------------------------------------------
    class_mapping           Required dictionary. Mapping from PascalVOC id to
                            its string label.
    ---------------------   -------------------------------------------
    chip_size               Optional integer. Size of the image to train the
                            model.
    ---------------------   -------------------------------------------
    val_split_pct           Optional float. Percentage of training data to keep
                            as validation.
    ---------------------   -------------------------------------------
    batch_size              Optional integer. Batch size for mini batch gradient
                            descent (Reduce it if getting CUDA Out of Memory
                            Errors).
    ---------------------   -------------------------------------------
    transforms              Optional tuple. Fast.ai transforms for data
                            augmentation of training and validation datasets
                            respectively (We have set good defaults which work
                            for satellite imagery well).
    ---------------------   -------------------------------------------
    collate_fn              Optional function. Passed to PyTorch to collate data
                            into batches(usually default works).
    ---------------------   -------------------------------------------
    seed                    Optional integer. Random seed for reproducible
                            train-validation split.
    =====================   ===========================================

    :returns: fastai DataBunch object
    """

    if not HAS_FASTAI:
        _raise_fastai_import_error()

    if type(path) is str:
        path = Path(path)

    images = 'images'

    get_y_func = partial(_get_bbox_lbls, class_mapping=class_mapping)

    src = (SSDObjectItemList.from_folder(path/images)
       .random_split_by_pct(val_split_pct, seed=seed)
       .label_from_func(get_y_func))

    if transforms is None:
        ranges = (0,1)
        train_tfms = [crop(size=chip_size, p=1., row_pct=ranges, col_pct=ranges), dihedral_affine()]
        val_tfms = [crop(size=chip_size, p=1., row_pct=0.5, col_pct=0.5)]
        transforms = (train_tfms, val_tfms)

    data = (src
        .transform(transforms, tfm_y=True)
        .databunch(bs=batch_size, collate_fn=collate_fn)
        .normalize(imagenet_stats)
       )

    data.chip_size = chip_size

    return data


import json
import datetime
import mimetypes
from urllib.parse import urlparse
from arcgis import env
from arcgis.gis import GIS
from arcgis.gis import Item
from ._ref import reference

class JournalStoryMap(object):
    """
    Represents a Journal Story Map

    ===============     ====================================================================
    **Argument**        **Description**
    ---------------     --------------------------------------------------------------------
    item                Optional Item. The storymap item.
    ---------------     --------------------------------------------------------------------
    gis                 Optional GIS. The connection to the Enterprise.
    ===============     ====================================================================

    """
    _properties = None
    _gis = None
    _itemid = None
    _item = None

    def __init__(self, item=None, gis=None):
        """initializer"""
        if gis is None:
            self._gis = env.active_gis
        else:
            self._gis = gis
        if item and isinstance(item, str):
            self._item = gis.content.get(item)
            self._itemid = self._item.itemid
            self._properties = self._item.get_data()
        elif item and isinstance(item, Item) and \
             'MapJournal' in item.typeKeywords:
            self._item = item
            self._itemid = self._item.itemid
            self._properties = self._item.get_data()
        elif item and isinstance(item, Item) and \
             'MapJournal' not in item.typeKeywords:
            raise ValueError("Item is not a Journal Story Map")
        else:
            self._properties = reference['journal']
    #----------------------------------------------------------------------
    def __str__(self):
        return json.dumps(self._properties)
    #----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()
    #----------------------------------------------------------------------
    def _refresh(self):
        if self._item:
            self._properties = json.loads(self._item.get_data())
    #----------------------------------------------------------------------
    @property
    def properties(self):
        """returns the storymap's JSON"""
        return self._properties
    #----------------------------------------------------------------------
    def add(self, title,
            url_or_item, content=None,
            actions=None, visible=True,
            alt_text="", display='stretch',
            **kwargs):
        """
        Adds a new section to the StoryMap

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        title               Required string. The title of the section.
        ---------------     --------------------------------------------------------------------
        url_or_item         Required string/Item. The web address to the resource or a Web Map
                            item.
        ---------------     --------------------------------------------------------------------
        content             Optional string. The content of the section.
        ---------------     --------------------------------------------------------------------
        actions             Optional list. A collection of actions performed on the section
        ---------------     --------------------------------------------------------------------
        visible             Optional boolean. If True, the section is visible on publish. If
                            False, the section is not displayed.
        ---------------     --------------------------------------------------------------------
        alt_text            Optional string. Specifies an alternate text for an image.
        ---------------     --------------------------------------------------------------------
        display             Optional string. The image display properties.
        ===============     ====================================================================


        **WebMap Options**

        ====================     ====================================================================
        **Argument**             **Description**
        --------------------     --------------------------------------------------------------------
        show_legend              Optional boolean. If True, the legend will be visible.
        --------------------     --------------------------------------------------------------------
        show_default_legend      Optional boolean. Shows the legend on default.
        --------------------     --------------------------------------------------------------------
        extent                   Optional dict/Envelope. The extent of the webmap.
        --------------------     --------------------------------------------------------------------
        layer_visibility         Optional list. The visibility of the layers in a webmap.  This is a
                                 list of dictionaries where the syntax is as follows:

                                 Syntax:

                                 [
                                    {
                                       "id" : "<id>",
                                       "visibility" : "<true/false>"
                                    }
                                 ]

                                 Example:

                                 [
                                 {
                                    "id" : "csv_6005_0",
                                    "visibility" : False,
                                 },
                                 {
                                    "id" : "csv_6006_0",
                                    "visibility" : True,
                                 }
                                 ]
        --------------------     --------------------------------------------------------------------
        popup                    Optional dict. The popup definition for the webmap.
        ====================     ====================================================================



        :return: Boolean


        """
        if isinstance(url_or_item, Item):
            show_legend = kwargs.pop("show_legend", False)
            show_default_legend = kwargs.pop("show_default_legend", False)
            extent = kwargs.pop("extent", None)
            layer_visibility = kwargs.pop("layer_visibility", None)
            popup = kwargs.pop("popup", None)
            if layer_visibility:
                layer_visibility = json.dumps(layer_visibility)
            return self._add_webmap(item=url_or_item, title=title, content=content,
                                    actions=actions, visible=visible, alt_text=alt_text,
                                    display=display,
                                    show_legend=show_legend,
                                    show_default_legend=show_default_legend,
                                    extent=extent,
                                    layer_visibility=layer_visibility,
                                    popup=popup)
        elif isinstance(url_or_item, str):
            mt = mimetypes.guess_type(url=url_or_item)
            if mt[0].lower().find('video') > -1:
                return self._add_video(url=url_or_item,
                                      title=title,
                                      content=content,
                                      actions=actions,
                                      visible=visible,
                                      alt_text=alt_text,
                                      display=display)
            elif mt[0].lower().find('image') > -1:
                return self._add_image(title=title, image=url_or_item,
                                       content=content, actions=actions, visible=visible,
                                       alt_text=alt_text, display=display)
            else:
                return self._add_webpage(title=title, url=url_or_item,
                                         content=content, actions=actions, visible=visible,
                                         alt_text=alt_text, display=display)
        return False
        #----------------------------------------------------------------------
    def _add_webpage(self,
                    title,
                    url,
                    content=None,
                    actions=None,
                    visible=True,
                    alt_text="",
                    display='stretch'):
        """
        Adds a webpage to the storymap

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        title               Required string. The title of the section.
        ---------------     --------------------------------------------------------------------
        url                 Required string. The web address of the webpage
        ---------------     --------------------------------------------------------------------
        content             Optional string. The content of the section.
        ---------------     --------------------------------------------------------------------
        actions             Optional list. A collection of actions performed on the section
        ---------------     --------------------------------------------------------------------
        visible             Optional boolean. If True, the section is visible on publish. If
                            False, the section is not displayed.
        ---------------     --------------------------------------------------------------------
        alt_text            Optional string. Specifies an alternate text for an image.
        ---------------     --------------------------------------------------------------------
        display             Optional string. The image display properties.
        ===============     ====================================================================


        :return: Boolean

        """
        if actions is None:
            actions = []
        if visible:
            visible = "PUBLISHED"
        else:
            visible = "HIDDEN"
        self._properties['values']['story']['sections'].append(
            {
                "title": title,
                "content": content,
                "contentActions": actions,
                "creaDate": int(datetime.datetime.now().timestamp() * 1000),
                "pubDate": int(datetime.datetime.now().timestamp() * 1000),
                "status": visible,
                "media": {
                    "type": "webpage",
                    "webpage": {
                        "url": url,
                        "type": "webpage",
                        "altText": alt_text,
                        "display": display,
                        "unload": True,
                        "hash": "5"
                    }
                }
            }
        )
        return True
    #----------------------------------------------------------------------
    def _add_video(self,
                  url,
                  title,
                  content,
                  actions=None,
                  visible=True,
                  alt_text="",
                  display='stretch'
                  ):
        """
        Adds a video section to the StoryMap.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        title               Required string. The title of the section.
        ---------------     --------------------------------------------------------------------
        url                 Required string. The web address of the image
        ---------------     --------------------------------------------------------------------
        content             Optional string. The content of the section.
        ---------------     --------------------------------------------------------------------
        actions             Optional list. A collection of actions performed on the section
        ---------------     --------------------------------------------------------------------
        visible             Optional boolean. If True, the section is visible on publish. If
                            False, the section is not displayed.
        ---------------     --------------------------------------------------------------------
        alt_text            Optional string. Specifies an alternate text for an image.
        ---------------     --------------------------------------------------------------------
        display             Optional string. The image display properties.
        ===============     ====================================================================


        :return: Boolean

        """
        if actions is None:
            actions = []
        if visible:
            visible = "PUBLISHED"
        else:
            visible = "HIDDEN"
        video = {
            "title": title,
            "content": content,
            "contentActions": actions,
            "creaDate": 1523450612336,
            "pubDate": 1523450580000,
            "status": visible,
            "media": {
                "type": "video",
                "video": {
                    "url": url,
                    "type": "video",
                    "altText": alt_text,
                    "display": display
                }
            }
        }
        self._properties['values']['story']['sections'].append(video)
        return True
    #----------------------------------------------------------------------
    def _add_webmap(self,
                   item,
                   title,
                   content,
                   actions=None,
                   visible=True,
                   alt_text="",
                   display='stretch',
                   show_legend=False,
                   show_default_legend=False,
                   extent=None,
                   layer_visibility=None,
                   popup=None
                   ):
        """
        Adds a WebMap to the Section.


        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        item                Required string/Item. The webmap Item Id or Item of a webmap.
        ---------------     --------------------------------------------------------------------
        title               Required string. The title of the section.
        ---------------     --------------------------------------------------------------------
        url                 Required string. The web address of the image
        ---------------     --------------------------------------------------------------------
        content             Optional string. The content of the section.
        ---------------     --------------------------------------------------------------------
        actions             Optional list. A collection of actions performed on the section
        ---------------     --------------------------------------------------------------------
        visible             Optional boolean. If True, the section is visible on publish. If
                            False, the section is not displayed.
        ---------------     --------------------------------------------------------------------
        alt_text            Optional string. Specifies an alternate text for an image.
        ---------------     --------------------------------------------------------------------
        display             Optional string. The image display properties.
        ===============     ====================================================================


        :return: Boolean

        """
        if isinstance(item, Item):
            item = item.itemid

        if actions is None:
            actions = []
        if visible:
            visible = "PUBLISHED"
        else:
            visible = "HIDDEN"
        wm = {
            "title": title,
            "content": content,
            "contentActions": actions,
            "creaDate": int(datetime.datetime.now().timestamp() * 1000),
            "pubDate": int(datetime.datetime.now().timestamp() * 1000),
            "status": visible,
            "media": {
                "type": "webmap",
                "webmap": {
                    "id": item,
                    "extent": extent,
                    "layers": layer_visibility,
                    "popup": popup,
                    "overview": {
                        "enable": False,
                        "openByDefault": True
                        },
                    "legend": {
                        "enable": show_legend,
                        "openByDefault": show_default_legend
                        },
                    "geocoder": {
                        "enable": False
                        },
                    "altText": alt_text
                }
            }
        }
        self._properties['values']['story']['sections'].append(wm)
        return True
    #----------------------------------------------------------------------
    def _add_image(self,
                  title,
                  image,
                  content=None,
                  actions=None,
                  visible=True,
                  alt_text=None,
                  display='fill'):
        """
        Adds a new image section to the storymap


        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        title               Required string. The title of the section.
        ---------------     --------------------------------------------------------------------
        url                 Required string. The web address of the image
        ---------------     --------------------------------------------------------------------
        content             Optional string. The content of the section.
        ---------------     --------------------------------------------------------------------
        actions             Optional list. A collection of actions performed on the section
        ---------------     --------------------------------------------------------------------
        visible             Optional boolean. If True, the section is visible on publish. If
                            False, the section is not displayed.
        ---------------     --------------------------------------------------------------------
        alt_text            Optional string. Specifies an alternate text for an image.
        ---------------     --------------------------------------------------------------------
        display             Optional string. The image display properties.
        ===============     ====================================================================


        :return: Boolean

        """
        if actions is None:
            actions = []
        if visible:
            visible = "PUBLISHED"
        else:
            visible = "HIDDEN"
        self._properties['values']['story']['sections'].append(
            {
                "title": title,
                "content": content,
                "contentActions": actions,
                "creaDate": int(datetime.datetime.now().timestamp() * 1000),
                "pubDate": int(datetime.datetime.now().timestamp() * 1000),
                "status": visible,
                "media": {
                    "type": "image",
                    "image": {
                        "url": image,
                        "type": "image",
                        "altText": alt_text,
                        "display": display
                    }
                }
            }
        )
        return True
    #----------------------------------------------------------------------
    def remove(self, index):
        """
        Removes a section by index.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        index               Required integer. The position of the section to remove.
        ===============     ====================================================================


        :return: Boolean

        """
        try:
            item = self._properties['values']['story']['sections'][index]
            self._properties['values']['story']['sections'].remove(item)
            return True
        except:
            return False
    #----------------------------------------------------------------------
    def save(self, title=None, tags=None, description=None):
        """
        Saves an Journal StoryMap to the GIS


        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        title               Optional string. The title of the StoryMap.
        ---------------     --------------------------------------------------------------------
        tags                Optional string. The tags of the StoryMap.
        ---------------     --------------------------------------------------------------------
        description         Optional string. The description of the StoryMap
        ===============     ====================================================================


        :return: Boolean

        """
        import uuid

        if self._item:
            p = {
                    'text' : json.dumps(self._properties)
                }
            if title:
                p['title'] = title
            if tags:
                p['tags'] = tags
            return self._item.update(item_properties=p)
        else:
            if title is None:
                title = "Map Journal, %s" % uuid.uuid4().hex[:10]
            if tags is None:
                tags = "Story Map,Map Journal"
            typeKeywords = ",".join(['JavaScript', 'layout-side', 'Map', 'MapJournal',
                                     'Mapping Site', 'Online Map', 'Ready To Use',
                                     'selfConfigured', 'Story Map', 'Story Maps',
                                     'Web Map'])
            item = self._gis.content.add(item_properties={
                'title' : title,
                'tags' : tags,
                'text' : json.dumps(self._properties),
                'typeKeywords' : typeKeywords,
                'itemType' : 'text',
                'type' : "Web Mapping Application",
            })
            parse = urlparse(self._gis._con.baseurl)
            isinstance(self._gis, GIS)
            if self._gis._portal.is_arcgisonline:
                url = "%s://%s/apps/MapJournal/index.html?appid=%s" % (parse.scheme, parse.netloc, item.itemid)
            else:
                import os
                wa = os.path.dirname(parse.path[1:])
                url = "%s://%s/%s/sharing/rest/apps/MapJournal/index.html?appid=%s" % (parse.scheme, parse.netloc, wa, item.itemid)
            return item.update(item_properties={
                'url' : url
            })
        return False
    #----------------------------------------------------------------------
    def delete(self):
        """Deletes the saved item on ArcGIS Online/Portal"""
        if self._item:
            return self._item.delete()
        return False
    #----------------------------------------------------------------------
    @property
    def panel(self):
        """
        Gets/Sets the panel state for the Journal Story Map
        """
        return self._properties["values"]["settings"]["layout"]["id"]
    #----------------------------------------------------------------------
    @panel.setter
    def panel(self, value):
        """
        Gets/Sets the panel state for the Journal Story Map
        """
        if value.lower() == "float":
            self._properties["values"]["settings"]["layout"]["id"] = "float"
        else:
            self._properties["values"]["settings"]["layout"]["id"] = "side"
    #----------------------------------------------------------------------
    @property
    def header(self):
        """gets/sets the headers for the Journal StoryMap"""
        default = {
            "social": {
                "bitly": True,
                "twitter": True,
                "facebook": True
                },
            "logoURL": None,
            "linkURL": "https://storymaps.arcgis.com",
            "logoTarget": "",
            "linkText": "A Story Map"
        }
        if 'header' in self._properties['values']['settings']:
            return self._properties['values']['settings']['header']
        else:
            self._properties['values']['settings']['header'] = default
            return default
    #----------------------------------------------------------------------
    @header.setter
    def header(self, value):
        """"""
        if value is None:
            default = {
                "social": {
                    "bitly": True,
                    "twitter": True,
                    "facebook": True
                    },
                "logoURL": None,
                "linkURL": "https://storymaps.arcgis.com",
                "logoTarget": "",
                "linkText": "A Story Map"
            }
            self._properties['values']['settings']['header'] = default
        else:
            self._properties['values']['settings']['header'] = value
    #----------------------------------------------------------------------
    @property
    def theme(self):
        """"""
        default = {
            "colors": {
                "text": "#FFFFFF",
                "name": "float-default-1",
                "softText": "#FFF",
                "media": "#a0a0a0",
                "themeMajor": "black",
                "panel": "#000000",
                "textLink": "#DDD",
                "esriLogo": "white",
                "dotNav": "#000000",
                "softBtn": "#AAA"
                },
            "fonts": {
                "sectionTitle": {
                    "value": "font-family:\'open_sansregular\', sans-serif;",
                    "id": "default"
                    },
                "sectionContent": {
                    "value": "font-family:\'open_sansregular\', sans-serif;",
                    "id": "default"
                }
            }
        }
        if 'theme' in self._properties['values']['settings']:
            return self._properties['values']['settings']['theme']
        else:
            self._properties['values']['settings']['theme'] = default
            return self._properties['values']['settings']['theme']
        return default
    #----------------------------------------------------------------------
    @theme.setter
    def theme(self, value):
        """"""
        default = {
            "colors": {
                "text": "#FFFFFF",
                "name": "float-default-1",
                "softText": "#FFF",
                "media": "#a0a0a0",
                "themeMajor": "black",
                "panel": "#000000",
                "textLink": "#DDD",
                "esriLogo": "white",
                "dotNav": "#000000",
                "softBtn": "#AAA"
                },
            "fonts": {
                "sectionTitle": {
                    "value": "font-family:\'open_sansregular\', sans-serif;",
                    "id": "default"
                    },
                "sectionContent": {
                    "value": "font-family:\'open_sansregular\', sans-serif;",
                    "id": "default"
                }
            }
        }
        if 'theme' in self._properties['values']['settings']:
            self._properties['values']['settings']['theme'] = value
        elif not 'theme' in self._properties['values']['settings']:
            self._properties['values']['settings']['theme'] = value
        elif value is None:
            self._properties['values']['settings']['theme'] = default








"""
Controls the local portal's security settings
"""
from ..._impl.connection import _ArcGISConnection
from .. import GIS
from ._base import BasePortalAdmin
########################################################################
class PasswordPolicy(BasePortalAdmin):
    """
    Manages a GIS Security Policy.  Administrators can view, update or
    reset the site's security policy.
    """
    _gis = None
    _con = None
    _url = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis=None, **kwargs):
        """Constructor"""
        super(PasswordPolicy, self).__init__(url=url,
                                             gis=gis,
                                             **kwargs)
        initialize = kwargs.pop("initialize", False)
        if isinstance(gis, _ArcGISConnection):
            self._con = gis
        elif isinstance(gis, GIS):
            self._gis = gis
            self._con = gis._con
        else:
            raise ValueError(
                "connection must be of type GIS or _ArcGISConnection")
        if initialize:
            self._init(self._gis)
    #----------------------------------------------------------------------
    def __str__(self):
        return '<%s at %s>' % (type(self).__name__, self._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return '<%s at %s>' % (type(self).__name__, self._url)
    #----------------------------------------------------------------------
    @property
    def policy(self):
        """gets/sets the current security policy"""
        if self._properties is None:
            self._init()
        return self._properties
    #----------------------------------------------------------------------
    @policy.setter
    def policy(self, value=None):
        """gets/sets the current security policy"""
        url = "%s/update" % self._url
        from ..._impl.common._mixins import PropertyMap
        if value is None:
            value = self.properties['passwordPolicy']
        if isinstance(value, (dict, PropertyMap)):
            value = dict(value)
        else:
            raise ValueError("Input must be a dictionary of PropertyMap")
        params = {"f" : "json",
                  }
        for k,v in value.items():
            params[k] = v
        res = self._con.post(url, params)
        if 'success' in res:
            self._properties = None
            return res['success']
        return res
    #----------------------------------------------------------------------
    def reset(self):
        """
        resets the security policy to the default install settings
        """
        url = "%s/reset" % self._url
        params = {'f' : 'json'}
        res = self._con.post(url, params)
        if 'success' in res:
            return res['success']
        return res
########################################################################
class Security(BasePortalAdmin):
    """
    This resource is an umbrella for a collection of system-wide resources
    for your portal. This resource provides access to the ArcGIS Web
    Adaptor configuration, portal directories, database management server,
    indexing capabilities, license information, and the properties of your
    portal.
    """
    _gis = None
    _con = None
    _url = None
    _oauth = None
    _eu = None
    _eg = None
    _ssl = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis=None, **kwargs):
        """Constructor"""
        super(Security, self).__init__(url=url,
                                       gis=gis,
                                       **kwargs)
        initialize = kwargs.pop("initialize", False)
        if isinstance(gis, _ArcGISConnection):
            self._con = gis
        elif isinstance(gis, GIS):
            self._gis = gis
            self._con = gis._con
        else:
            raise ValueError(
                "connection must be of type GIS or _ArcGISConnection")
        if initialize:
            self._init(self._gis)
    #----------------------------------------------------------------------
    @property
    def enterpriseusers(self):
        """
        provides access into managing enterprise users
        """
        if self._eu is None:
            url = "%s/users" % self._url
            self._eu = EnterpriseUsers(url=url, gis=self._gis)
        return self._eu
    #----------------------------------------------------------------------
    @property
    def groups(self):
        """provides access to managing Enterprise Groups with Portal"""
        if self._eg is None:
            url = "%s/groups" % self._url
            self._eg = EnterpriseGroups(url=url, gis=self._gis)
        return self._eg
    #----------------------------------------------------------------------
    @property
    def tokens(self):
        """
        This resource represents the token configuration within your
        portal. Use the set on token_config operation to change the
        configuration properties of the token service.
        """
        url = "%s/tokens" % self._url
        params = {"f" : "json"}
        return self._con.get(path=url, params=params)
    #----------------------------------------------------------------------
    @tokens.setter
    def tokens(self, value):
        """
        This resource represents the token configuration within your
        portal. Use the set on token_config operation to change the
        configuration properties of the token service.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        value                           Required string. A shared key value
        ===========================     ====================================================================

        :returns: dict

        """
        import six
        params = {
            "f" : "json",
            "tokenConfig" : None}
        if isinstance(value, six.string_types):
            params['tokenConfig'] = {"sharedKey": value}
        elif isinstance(value, dict) and \
             "sharedKey" in value:
            params['tokenConfig'] = value
        else:
            raise ValueError("invalid value given")
        url = "%s/tokens/update" % self._url
        return self._con.post(path=url,
                              postdata=params)
    #----------------------------------------------------------------------
    @property
    def oauth(self):
        """
        The OAuth resource contains a set of operations that update the
        OAuth2-specific properties of registered applications in Portal for
        ArcGIS.
        """
        if self._oauth is None:
            url = "%s/oauth" % self._url
            self._oauth = OAuth(url=url, gis=self._gis)
        return self._oauth
    #----------------------------------------------------------------------
    @property
    def config(self):
        """
        The security configuration consists of the identity store
        configuration.
        If your portal will be authenticated through ArcGIS Web Adaptor,
        you must set up your preferred authentication on your web server.
        Use the Update Identity Store operation to configure your portal to
        connect to your enterprise identity provider such as Windows Domain
        or LDAP. By default, Portal for ArcGIS is configured to use the
        built-in store and token-based authentication.
        """
        url = "%s/config" % self._url
        params = {"f" : "json"}
        return self._con.get(path=url, params=params)
    #----------------------------------------------------------------------
    @config.setter
    def config(self, value):
        """
        This operation can be used to update the portal's security settings
        such as whether or not enterprise accounts are automatically
        registered as members of your ArcGIS organization the first time
        they accesses the portal.
        The security configuration is stored as a collection of properties
        in a JSON object. The following properties are supported:
         - enableAutomaticAccountCreation
         - disableServicesDirectory
         - defaultRoleForUser (introduced at ArcGIS 10.4)
        The automatic account creation flag (enableAutomaticAccountCreation)
        determines the behavior for unregistered enterprise accounts the
        first time they access the portal. When the value for this property
        is set to false, first time users are not automatically registered
        as members of your ArcGIS organization, and have the same access
        privileges as other nonmembers. For these accounts to sign in, an
        administrator must register the enterprise accounts using the
        Create User operation.
        The default value for the enableAutomaticAccountCreation property
        is false. When this value is set to true, portal will add
        enterprise accounts automatically as members of your ArcGIS
        organization.
        The disableServicesDirectory property controls whether the HTML
        pages of the services directory should be accessible to the users.
        The default value for this property is false, meaning the services
        directory HTML pages are accessible to everyone.
        Use the defaultRoleForUser property to set which role the portal
        automatically assigns to new member accounts. By default, new
        accounts are assigned to account_user. Other possible values are
        account_publisher or the ID of one of the custom roles defined in
        the ArcGIS organization. To obtain the ID of a custom role,
         - Log in to the portal sharing directory.
         - Go to Portals > Self > Roles.
         - Copy the custom role ID you want to use.
        The allowedProxyHosts property restricts what hosts the portal can
        access directly. This restriction applies to several scenarios,
        including when the portal accesses resources from a server that
        does not support Cross Origin Resource Sharing (CORS) or when
        saving credentials used to access a secure service. By default,
        this property is not defined and no restrictions are applied.
        Define the allowedProxyHosts with a comma-separated list of
        hostnames to restrict the hosts the portal can access directly. Use
        the format (.*).domain.com to allow access to all machines within a
        specified domain.

        *example value*
          {
           "disableServicesDirectory":false,
           "enableAutomaticAccountCreation":true,
           "defaultRoleForUser": 12aBC3D4EF5ghIJ
          }
        """
        url = "%s/config/update" % self._url
        params = {
            "securityConfig" : value,
            "f" : "json"
        }
        return self._con.post(path=url, postdata=params)
    #----------------------------------------------------------------------
    def update_identity_store(self,
                              user_config=None,
                              group_config=None):
        """
        You can use this operation to change the identity provider and
        group store configuration in your portal. When Portal for ArcGIS is
        first installed, it supports token-based authentication and
        built-in groups using the built-in identity store for accounts. To
        configure your portal to connect to your enterprise authentication
        mechanism and group store, it must be configured to use an
        enterprise identity store such as Windows Active Directory or LDAP.

        See: http://resources.arcgis.com/en/help/arcgis-rest-api/index.html#/Update_Identity_Store/02r300000249000000/

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        user_config                     Optional dict. The user store configuration
        ---------------------------     --------------------------------------------------------------------
        group_config                    Optional dict. The group store configuration
        ===========================     ====================================================================

        :returns: dict

        """
        url = "%s/config/updateIdentityStore" % self._url
        if user_config is None:
            user_config = {
                "type": "BUILTIN",
                "properties": {"isPasswordEncrypted": "true"}
            }
        if group_config is None:
            group_config = {
                "type": "BUILTIN",
                "properties": {"isPasswordEncrypted": "true"}
            }
        params = {
            "userStoreConfig": user_config,
            "groupStoreConfig": group_config,
            "f" : "json"
        }
        return self._con.post(path=url, postdata=params)
    #----------------------------------------------------------------------
    @property
    def test_identity_store(self,
                            user_config=None,
                            group_config=None):
        """
        This operation can be used to test the connection to a user or
        group store.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        user_config                     Optional dict. The user store configuration
        ---------------------------     --------------------------------------------------------------------
        group_config                    Optional dict. The group store configuration
        ===========================     ====================================================================

        :returns: dict


        """
        if user_config is None and \
           group_config is None:
            return
        params = {
            "f" : "json"
        }
        url = "%s/config/testIdentityStore" % self._url
        if user_config is not None:
            user_config = {
                "type": "BUILTIN",
                "properties": {"isPasswordEncrypted": "true"}
            }
            params["userStoreConfig"] = user_config
        if group_config is not None:
            group_config = {
                "type": "BUILTIN",
                "properties": {"isPasswordEncrypted": "true"}
            }
            params["groupStoreConfig"] = group_config

        return self._con.post(path=url, postdata=params)
    #----------------------------------------------------------------------
    @property
    def ssl(self):
        """
        Provides access to managing and updating SSL Certificates on a
        Portal site.
        """
        if self._ssl is None:
            url = "%s/sslCertificates" % self._url
            self._ssl = SSLCertificates(url=url, gis=self._gis)
        return self._ssl
########################################################################
class OAuth(BasePortalAdmin):
    """
    The OAuth resource contains a set of operations that update the
    OAuth2-specific properties of registered applications in Portal for
    ArcGIS.
    """
    _gis = None
    _con = None
    _url = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis=None, **kwargs):
        """Constructor"""
        super(OAuth, self).__init__(url=url,
                                    gis=gis,
                                    **kwargs)
        initialize = kwargs.pop("initialize", False)
        if isinstance(gis, _ArcGISConnection):
            self._con = gis
        elif isinstance(gis, GIS):
            self._gis = gis
            self._con = gis._con
        else:
            raise ValueError(
                "connection must be of type GIS or _ArcGISConnection")
        if initialize:
            self._init(self._gis)
    #----------------------------------------------------------------------
    def update(self, current_id, new_id):
        """
        When new applications are registered with Portal for ArcGIS, a new
        client ID is generated for the application. This allows the
        application to access content from the portal. The new client ID
        does not work if the application developer has programmed against a
        specific ID. This operation can, therefore, be used to change the
        client ID to another value as specified by the application
        developer.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        current_id                      Required string. The current client ID of an existing application.
        ---------------------------     --------------------------------------------------------------------
        new_id                          Required string. The new client ID to assign to the application.
        ===========================     ====================================================================

        :returns: dict

        """
        params = {"f" : "json",
                  "currentAppID" : current_id,
                  "newAppID" : new_id}
        url = "%s/changeAppID" % self._url
        res = self._con.post(path=url, postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        elif 'success' in res:
            return res['success']
        return res
    #----------------------------------------------------------------------
    @property
    def app_info(self):
        """
        Every application registered with Portal for ArcGIS has a unique
        client ID and a list of redirect URIs that are used for OAuth. This
        operation returns these OAuth-specific properties of an
        application. You can use this information to update the redirect
        URIs by using the Update App Info operation.
        """
        url = "%s/getAppInfo" % self._url
        params = {"f" : "json"}
        return self._con.get(path=url, params=params)
    #----------------------------------------------------------------------
    @app_info.setter
    def app_info(self, value):
        """
        This operation allows you to update the OAuth-specific properties
        associated with an application. Use the Get App Info operation to
        obtain the existing OAuth properties that can be edited.
        """
        url = "%s/getAppInfo" % self._url
        params = {"f" : "json",
                  "appInfo" : value}
        return self._con.post(path=url, postdata=params)
########################################################################
class SSLCertificates(BasePortalAdmin):
    """
    Manages the Portal's SSL Certificates
    """
    _gis = None
    _con = None
    _url = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis=None, **kwargs):
        """Constructor"""
        self._certs = None
        super(SSLCertificates, self).__init__(url=url,
                                              gis=gis,
                                              **kwargs)

        initialize = kwargs.pop("initialize", False)
        if isinstance(gis, _ArcGISConnection):
            self._con = gis
        elif isinstance(gis, GIS):
            self._gis = gis
            self._con = gis._con
        else:
            raise ValueError(
                "connection must be of type GIS or _ArcGISConnection")
        if initialize:
            self._init(self._gis)
    #----------------------------------------------------------------------
    def _refresh(self):
        """reloads all the properties of a given service"""
        self._init()
    #----------------------------------------------------------------------
    def update(self, alias, protocols, cipher_suites):
        """
        Use this operation to configure the web server certificate, SSL
        protocols, and cipher suites used by the portal.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        alias                           Required string. The name of the certificate. This is a required
                                        parameter. The certificate must be already present in the portal.
        ---------------------------     --------------------------------------------------------------------
        protocols                       Required string. The SSL protocols the portal will use. Valid
                                        options are TLSv1, TLSv1.1, and TLSv1.2; values must be comma
                                        separated. By default, these options are all enabled.
        ---------------------------     --------------------------------------------------------------------
        cipher_suites                   Required string. The cipher suites the portal will use. Valid
                                        options are:
                                            - TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256
                                            - TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256
                                            - TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA
                                            - TLS_ECDHE_RSA_WITH_3DES_EDE_CBC_SHA
                                            - TLS_RSA_WITH_AES_128_GCM_SHA256
                                            - TLS_RSA_WITH_AES_128_CBC_SHA256
                                            - TLS_RSA_WITH_AES_128_CBC_SHA
                                            - TLS_RSA_WITH_3DES_EDE_CBC_SHA
                                        By default, all of the above options are enabled. Values must be
                                        comma separated.
        ===========================     ====================================================================

        :returns: dict

        """
        self._certs = None
        url = "%s/update" % self._url
        params = {
            "f" : "json",
            "webServerCertificateAlias" : alias,
            "sslProtocols" : protocols,
            "cipherSuites" : cipher_suites
        }
        return self._con.post(path=url,
                              postdata=params)
    #----------------------------------------------------------------------
    def generate(self,
                 alias,
                 common_name,
                 organization,
                 key_algorithm="RSA",
                 validity=90,
                 key_size=2048,
                 signature_algorithm="SHA256withRSA",
                 unit="",
                 city="",
                 state="",
                 country_code="",
                 alt_name=""):
        """
        Use this operation to create a self-signed certificate or as a
        starting point for getting a production-ready CA-signed
        certificate. The portal will generate a certificate for you and
        store it in its keystore.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        alias                           Required string. The name of the certificate. This is a required
                                        parameter.
        ---------------------------     --------------------------------------------------------------------
        common_name                     Required string. The common name used to identify the server for
                                        which the certificate is to be generated. This is a required
                                        parameter.
        ---------------------------     --------------------------------------------------------------------
        organization                    Required string. The name of the organization. This is a required
                                        parameter.
        ---------------------------     --------------------------------------------------------------------
        key_algorithm                   Optional string. The algorithm used to generate the key pairs. The
                                        default is RSA.
        ---------------------------     --------------------------------------------------------------------
        validity                        Optional integer. The expiration time for the certificate in days.
                                        The default is 90.
        ---------------------------     --------------------------------------------------------------------
        key_size                        Optional integer. The size of the key. The default is 2048.
        ---------------------------     --------------------------------------------------------------------
        signature_algorithm             Optional string. The algorithm used to sign the self-signed
                                        certificates. The default is derived from the key_algorithm parameter.
        ---------------------------     --------------------------------------------------------------------
        unit                            Optional string. The department within which this server resides.
        ---------------------------     --------------------------------------------------------------------
        city                            Optional string. The name of the city
        ---------------------------     --------------------------------------------------------------------
        state                           Optional string. The name of the state
        ---------------------------     --------------------------------------------------------------------
        country_code                    Optional string. The two letter abbrevation of the country
        ---------------------------     --------------------------------------------------------------------
        alt_name                        Optional string. The common name used to identify the server for
                                        which the certificate is to be generated. This is a required
                                        parameter.
        ===========================     ====================================================================

        :returns: boolean

        """
        import json
        params = {
            "alias":alias,
            'keyAlg':key_algorithm,
            'keySize': key_size,
            'sigAlg':signature_algorithm,
            'cn':common_name.upper(),
            'orgUnit' : unit,
            'org':organization,
            'city': city,
            'state': state,
            'country':country_code,
            'validity': validity,
            'san':alt_name
        }
        self._certs = None
        url = "%s/generateCertificate" % self._url
        try:
            res = self._con.post(path=url, postdata=params)
        except json.JSONDecodeError:
            # Need to capture this because method only returns HTML
            # Ignore decoding errors
            self._refresh()
            return True
        except:
            return False

    #----------------------------------------------------------------------
    def import_certificate(self, certificate, alias, norestart=False):
        """
        This operation imports a certificate authority's (CA) root and
        intermediate certificates into the keystore.
        To create a production quality CA-signed certificate, you need
        to add the CA certificates into the keystore that enables the
        SSL mechanism to trust the CA (and the certificates it has
        signed). While most of the popular CA certificates are already
        available in the keystore, you can use this operation if you
        have a custom CA or specific intermediate certificates.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        certificate                     Required string. The file location of the certificate file
        ---------------------------     --------------------------------------------------------------------
        alias                           Required string. The name of the certificate
        ---------------------------     --------------------------------------------------------------------
        norestart                       Optional boolean. Determines if the portal should be prevented from
                                        restarting after importing the certificate. By default this is false
                                        and the portal will restart.  Added in 10.6.
        ===========================     ====================================================================

        :returns: boolean

        .. code-block:: python

            USAGE: Import a trusted CA or Intermediate SSL Certificate into Portal Admin API

            from arcgis.gis import GIS
            gis = GIS("https://yourportal.com/portal", "portaladmin", "password")
            # Get the SSL Certificate class
            sslmgr = gis.admin.security.ssl
            # Load a trust CA certificate and restart Portal
            resp = sslmgr.import_certificate(r'c:\\temp\\myTrustedCA.crt', 'myroot', norestart=False)
            print(resp)

            # Output
            True

        """

        from six.moves.urllib.error import HTTPError

        self._certs = None
        params = {
            "alias" : alias,
            "norestart" : norestart,
            "f" : "json"
        }
        files = {
            'file' : certificate
        }
        url = "%s/importRootOrIntermediate" % self._url
        try:
            res = self._con.post(path=url,
                                 add_headers=[('Accept', "*/*"), ('Accept-Encoding', 'gzip, deflate'),
                                              ('User-Agent', 'geosaurus/1.0'), ('Connection', 'keep-alive')],
                                 postdata=params,
                                 files=files)
        except HTTPError as error:
            if error.code == '408' or error.code == 408:
                return True
            return False
        except:
            return True
        return True
    #----------------------------------------------------------------------
    def import_server_certificate(self,
                                  alias,
                                  password,
                                  certificate):
        """
        This operation imports an existing server certificate, stored
        in the PKCS #12 format, into the keystore. If the certificate
        is a CA signed certificate, you must first import the CA Root
        or Intermediate certificate using the Import Root or
        Intermediate Certificate operation.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        alias                           Required string. The name of the certificate
        ---------------------------     --------------------------------------------------------------------
        password                        Required string. The password for the certificate
        ---------------------------     --------------------------------------------------------------------
        certificate                     Required string. The file location of the certificate file
        ===========================     ====================================================================

        :returns: boolean

        """
        params = {
            "f" : "json",
            "password" : password,
            "alias" : alias
        }
        files = {
            "file" : certificate
        }
        url = "%s/importExistingServerCertificate" % self._url
        try:
            return self._con.post(path=url,
                                  postdata=params,
                                  files=files)
        except:
            return False
        return True
    #----------------------------------------------------------------------
    def list(self, force=False):
        """
        List of SSL Certificates as represented in the Portal Admin API

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        force                           Optional Boolean. If True, the certificate list will be refreshed,
                                        else, if a set of values is in memory, it will use those values.
                                        This is used when you want to ensure you have the most up to date
                                        list of certificates.
        ===========================     ====================================================================

        :returns:
            List of SSLCertificate objects

        .. code-block:: python

            USAGE: Print out information about each SSL Certificate

            from arcgis.gis import GIS
            gis = GIS("https://yourportal.com/portal", "portaladmin", "password")
            # Get the SSL Certificate class
            sslmgr = gis.admin.security.ssl
            # Get a list of SSL Certificates
            sslcerts = sslmgr.list()
            # For each certificate, print its alias and issuer
            for sslcert in sslcerts:
                print("{} : {}".format(dict(sslcert)['aliasName'], dict(sslcert)['issuer']))

            # Output
            portal : CN=YOURPORTAL.COM, OU=Self Signed Certificate
            yourorgroot : CN=YourOrg Enterprise Root, DC=empty, DC=local
            samlcert : CN=YOURPORTAL.COM, OU=Self Signed Certificate
            ca_signed : CN=YourOrg Enterprise Root, DC=empty, DC=local

        """

        certs = []
        if self._certs is None or force:
            self._refresh()
            for cert in self.properties.sslCertificates:
                url = "%s/%s" % (self._url, cert)
                certs.append(SSLCertificate(url=url, gis=self._gis))
                del cert
            self._certs = certs
        return self._certs
    #----------------------------------------------------------------------
    def get(self, alias_name):
        """
        gets a single SSLCertificate object by the alias name

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        alias_name                      Required string. The common name of the certificate.
        ===========================     ====================================================================

        :returns: SSLCertificate Object

        .. code-block:: python

            USAGE: Print out information about a specific SSL Certificate by alias name

            from arcgis.gis import GIS
            gis = GIS("https://yourportal.com/portal", "portaladmin", "password")
            # Get the SSL Certificate class
            sslmgr = gis.admin.security.ssl
            # Get a specific certificate alias and print information
            ssl = sslmgr.get('portal')
            for prop in ssl.properties:
                print(prop, ssl.properties[prop])]))

            # Output
            aliasName portal
            issuer CN=YOURPORTAL.COM, OU=Self Signed Certificate
            subject CN=YOURPORTAL.COM, OU=Self Signed Certificate
            subjectAlternativeNames []
            validFrom Fri Sep 15 07:46:45 EDT 2017
            validUntil Sun Jul 24 07:46:45 EDT 2050
            keyAlgorithm RSA
            keySize 2048
            serialNumber 503b23c6
            version 3
            signatureAlgorithm SHA256withRSA
            keyUsage []
            md5Fingerprint 76d695d72e46b30ea90013676d559faa
            sha1Fingerprint 6f36513757c28ad43c2df5e4c7cee581ad18dd1e
            sha256Fingerprint a051aab19d1ed8ceee7322572b3b1b2abd1ed680d0a1d81d0da84cf0e1a1b6cb

        """
        for cert in self.list():
            if cert.properties['aliasName'].lower() == alias_name.lower():
                return cert
            del cert
        return None
########################################################################
class SSLCertificate(BasePortalAdmin):
    """
    represents a single registered certificate
    """
    _gis = None
    _con = None
    _url = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis=None, **kwargs):
        """Constructor"""
        super(SSLCertificate, self).__init__(url=url,
                                             gis=gis,
                                             **kwargs)
        initialize = kwargs.pop("initialize", False)
        if isinstance(gis, _ArcGISConnection):
            self._con = gis
        elif isinstance(gis, GIS):
            self._gis = gis
            self._con = gis._con
        else:
            raise ValueError(
                "connection must be of type GIS or _ArcGISConnection")
        if initialize:
            self._init(self._gis)
    #----------------------------------------------------------------------
    def generate_csr(self):
        """
        This operation generates a certificate signing request (CSR) for a
        self-signed certificate. A CSR is required by a CA to create a
        digitally signed version of your certificate.

        :returns: string

        """
        params = {
            "f" : "json"
        }
        url = "%s/generateCSR" % self._url
        res = self._con.post(path=url, postdata=params)
        if "certificateSigningRequest" in res:
            return res["certificateSigningRequest"]
        return res
    #----------------------------------------------------------------------
    def export(self, out_path=None):
        """
        This operation downloads an SSL certificate. The file returned by
        the server is an X.509 certificate. The downloaded certificate can
        be imported into a client that is making HTTP requests.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        out_path                        Required string. Save location of the certificate
        ===========================     ====================================================================

        :returns: string

        """
        if out_path is None:
            import tempfile
            out_path = tempfile.gettempdir()
        url = "%s/export" % self._url
        return self._con.get(path=url,
                             params={"f" : "json"},
                             out_folder=out_path)
    #----------------------------------------------------------------------
    def delete(self):
        """
        This operation deletes an SSL certificate from the key store. Once
        a certificate is deleted, it cannot be retrieved or used to enable
        SSL.
        """
        import json
        params = {"f" : "json"}
        url = "%s/delete" % self._url
        try:
            self._con.post(path=url, postdata=params)
            return True
        except json.JSONDecodeError:
            return True
        except:
            return False
    #----------------------------------------------------------------------
    def import_signed_certificate(self, file_path):
        """
        imports a certificate authority (CA) signed SSL certificate into
        the key store.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        file_path                       Required string. The location of the certificate
        ===========================     ====================================================================

        :returns: dict

        """
        url = "%s/importSignedCertificate" % self._url
        params = {'f' : "json"}
        return self._con.post(path=url,
                              postdata=params,
                              files={"file" : file_path})
########################################################################
class EnterpriseGroups(BasePortalAdmin):
    """
    The groups resource is an umbrella for operations to manage
    enterprise groups within the portal. The resource returns the total
    number of groups in the system.
    """
    _gis = None
    _con = None
    _url = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis=None, **kwargs):
        """Constructor"""
        super(EnterpriseGroups, self).__init__(url=url,
                                               gis=gis,
                                               **kwargs)
        initialize = kwargs.pop("initialize", False)
        if isinstance(gis, _ArcGISConnection):
            self._con = gis
        elif isinstance(gis, GIS):
            self._gis = gis
            self._con = gis._con
        else:
            raise ValueError(
                "connection must be of type GIS or _ArcGISConnection")
        if initialize:
            self._init(self._gis)
    #----------------------------------------------------------------------
    def search(self,
               query="",
               max_count=255):
        """
        This operation searches groups in the configured enterprise group
        store. You can narrow down the search using the filter parameter.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        query                           Optional string. Where clause into parse down results
        ---------------------------     --------------------------------------------------------------------
        max_count                       Optional integer. The maximum number of records to return
        ===========================     ====================================================================

        :returns: dict

        """
        params = {
            "f" : "json",
            "filter" : query,
            "maxCount" : max_count
        }
        url = "%s/searchEnterpriseGroups" % self._url
        return self._con.post(path=url, postdata=params)
    #----------------------------------------------------------------------
    def refresh_groups(self, groups):
        """
        This operation iterates over every enterprise account configured in
        the portal and determines if the user account is a part of the
        input enterprise group. If there are any change in memberships, the
        database and the indexes are updated for each group.
        While portal automatically refreshes the memberships during a user
        login and during a periodic refresh configured through the Update
        Identity Store operation, this operation allows an administrator to
        force a refresh.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        groups                          Required string. The comma seperated list of group names to be
                                        refreshed
        ===========================     ====================================================================

        :returns: dict

        """
        url = "%s/refreshMembership" % self._url
        params = {
            "f" : "json",
            "groups" : groups
        }
        return self._con.post(path=url, postdata=params)
    #----------------------------------------------------------------------
    def get_group_users(self, name, query="", max_count=255):
        """
        This operation returns the users that are currently assigned to the
        enterprise group within the enterprise user/group store. You can
        use the filter parameter to narrow down the user search.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        name                            Optional string. The name of the enterprise group
        ---------------------------     --------------------------------------------------------------------
        query                           Optional string. Where clause into parse down results
        ---------------------------     --------------------------------------------------------------------
        max_count                       Optional integer. The maximum number of records to return
        ===========================     ====================================================================

        :returns: dict

        """
        url = "%s/getUsersWithinEnterpriseGroup" % self._url
        params = {
            "f" : "json",
            "groupName" : name,
            "filter" : query,
            "maxCount" : max_count
        }
        return self._con.get(path=url, params=params)
    #----------------------------------------------------------------------
    def get_user_groups(self, username,
                        query="", max_count=255):
        """
        This operation lists the groups assigned to a user account in the
        configured enterprise group store.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        username                        Optional string. The name of the user account
        ---------------------------     --------------------------------------------------------------------
        query                           Optional string. Where clause into parse down results
        ---------------------------     --------------------------------------------------------------------
        max_count                       Optional integer. The maximum number of records to return
        ===========================     ====================================================================

        :returns: dict

        """
        url = "%s/getEnterpriseGroupsForUser" % self._url
        params = {
            "f" : "json",
            "username" : username,
            "filter" : query,
            "maxCount" : max_count
        }
        return self._con.get(path=url, params=params)
########################################################################
class EnterpriseUsers(BasePortalAdmin):
    """
    The users resource is an umbrella for operations to manage members
    within Portal for ArcGIS. The resource returns the total number of
    members in the system.
    """
    _gis = None
    _con = None
    _url = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis=None, **kwargs):
        """Constructor"""
        super(EnterpriseUsers, self).__init__(url=url,
                                              gis=gis,
                                              **kwargs)
        initialize = kwargs.pop("initialize", False)
        if isinstance(gis, _ArcGISConnection):
            self._con = gis
        elif isinstance(gis, GIS):
            self._gis = gis
            self._con = gis._con
        else:
            raise ValueError(
                "connection must be of type GIS or _ArcGISConnection")
        if initialize:
            self._init(self._gis)
    #----------------------------------------------------------------------
    def create(self,
               username,
               password,
               first_name,
               last_name,
               email,
               role="org_user",
               level=2,
               provider="arcgis",
               idp_username=None,
               description=None):
        """
        This operation is used to pre-create built-in or enterprise
        accounts within the portal. The provider parameter is used to
        indicate the type of user account.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        username                        Required string. The name of the user account
        ---------------------------     --------------------------------------------------------------------
        password                        Required string. The password of the user account
        ---------------------------     --------------------------------------------------------------------
        first_name                      Required string. The first name for the account
        ---------------------------     --------------------------------------------------------------------
        last_name                       Required string. The last name for the account
        ---------------------------     --------------------------------------------------------------------
        email                           Required string. The email for the account
        ---------------------------     --------------------------------------------------------------------
        role                            Optional string. The role for the user account. The default value is
                                        org_user.
                                        Values org_user | org_publisher | org_admin
        ---------------------------     --------------------------------------------------------------------
        level                           Optional integer. The account level to assign the user.
                                        Values 1 or 2
        ---------------------------     --------------------------------------------------------------------
        provider                        Optional string. The provider for the account. The default value is
                                        arcgis. Values arcgis | enterprise
        ---------------------------     --------------------------------------------------------------------
        idp_username                    Optional string. The name of the user as stored by the enterprise
                                        user store. This parameter is only required if the provider
                                        parameter is enterprise.
        ---------------------------     --------------------------------------------------------------------
        description                     Optional string. A user description
        ===========================     ====================================================================

        :returns: boolean

        """
        url = "%s/createUser" % self._url
        params = {
            "f" : "json",
            "username" : username,
            "password" : password,
            "firstname" : first_name,
            "lastname" : last_name,
            "email" : email,
            "role" : role,
            "level" : level,
            "provider" : provider
        }
        if idp_username:
            params['idpUsername'] =  idp_username
        if description:
            params['description'] = description
        res = self._con.post(path=url, postdata=params)
        return res['status'] == 'success'
    #----------------------------------------------------------------------
    def get(self, username):
        """
        This operation returns the description, full name, and email
        address for a single user in the enterprise identity (user) store
        configured with the portal. The username parameter is used to
        specify the enterprise username. If the user does not exist, an
        error is returned.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        username                        Required string. Username of the enterprise account. For Windows
                                        Active Directory users, this can be either domain\\username or just
                                        username. For LDAP users, the format is always username.
        ===========================     ====================================================================

        :returns: dict

        """
        url = "%s/getEnterpriseUsers" % self._url
        params = {
            "f" : "json",
            "username" : username
        }
        return self._con.get(path=url,
                             params=params)
    #----------------------------------------------------------------------
    def update(self, username, idp_username):
        """
        This operation allows an administrator to update the idp_username
        for an enterprise user in the portal. This is used when migrating
        from accounts used with web-tier authentication to SAML
        authentication.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        username                        Required string. Username of the enterprise account. For Windows
                                        Active Directory users, this can be either domain\\username or just
                                        username. For LDAP users, the format is always username.
        ---------------------------     --------------------------------------------------------------------
        idp_username                    Required string. The username used by the SAML identity provider
        ===========================     ====================================================================

        :returns: dict

        """
        url = "%s/updateEnterpriseUser" % self._url
        params = {
            "f" : "json",
            "username" : username,
            "idpUsername" : idp_username
        }
        res = self._con.get(path=url, params=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    def search(self, query="", max_count=255):
        """
        This operation searches users in the configured enterprise user
        store. You can narrow down the search using the filter parameter.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        query                           Optional string. Where clause into parse down results
        ---------------------------     --------------------------------------------------------------------
        max_count                       Optional integer. The maximum number of records to return
        ===========================     ====================================================================

        :returns: dict

        """
        url = "%s/searchEnterpriseUsers" % self._url
        params = {
            "f" : "json",
            "filter" : query,
            "maxCount" : max_count
        }
        return self._con.get(path=url, params=params)
    #----------------------------------------------------------------------
    def refresh_users(self, users):
        """
        This operation iterates over every enterprise group configured in
        the portal and determines if the input user accounts belong to any
        of the configured enterprise groups. If there is any change in
        membership, the database and the indexes are updated for each user
        account. While portal automatically refreshes the memberships
        during a user login and during a periodic refresh (configured
        through the Update Identity Store operation), this operation allows
        an administrator to force a refresh.

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        users                           Required string. A comma seperated list of users.
        ===========================     ====================================================================

        :returns: dict

        """
        params = {
            "f" : "json",
            "users" : users
        }
        url = "%s/refreshMembership" % self._url
        return self._con.post(path=url, postdata=params)



"""
This resource lists all the portal machines in a site. Each portal machine
has a status that indicates whether the machine is ready to accept
requests.
"""
from ._base import BasePortalAdmin
from ..._impl.common._mixins import PropertyMap
########################################################################
class Machines(BasePortalAdmin):
    """
    This resource lists all the portal machines in a site. Each portal machine
    has a status that indicates whether the machine is ready to accept
    requests.
    """
    _pa = None
    _url = None
    _properties = None
    _json_dict = None
    _json = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis, portaladmin, **kwargs):
        """Constructor"""
        super(Machines, self).__init__(url=url,
                                       gis=gis,
                                       portaladmin=portaladmin)
        initialize = kwargs.pop('initialize', False)
        self._url = url
        self._pa = portaladmin
        if initialize:
            self._init()
    #----------------------------------------------------------------------
    def list(self):
        """
        provides a list of all registered machines with the local GIS
        """
        machines = []
        for m in self.properties.machines:
            machines.append(Machine(name=m['machineName'],
                                    url=self._url,
                                    info=dict(m),
                                    gis=self._gis,
                                    portaladmin=self._pa))
        return machines
    #----------------------------------------------------------------------
    def get(self, name):
        """
        allows for retrieval of a single instance of Machine by it's
        registered name.
        """
        for m in self.properties.machines:
            if m['machineName'].lower() == name.lower():
                return Machine(name=m['machineName'],
                               gis=self._gis,
                               url=self._url,
                               info=dict(m),
                               portaladmin=self._pa)
        return
########################################################################
class Machine(BasePortalAdmin):
    """
    Represents a single machine instance registered with the GIS

    Parameters:
    :param name: machine name
    :param url: base url for the machine
    :param info: dictionary representing a single machine
    :param portaladmin: PortalAdminManager object
    :param initialize: (optional) if True, properties are loaded on
    creation
    """
    _pa = None
    _con = None
    _info = None
    _name = None
    _url = None
    _properties = None
    _json_dict = None
    _json = None
    #----------------------------------------------------------------------
    def __init__(self, name,
                 url, info,
                 gis, portaladmin,
                 **kwargs):
        """Constructor"""
        self._info = info
        super(Machine, self).__init__(url=url,
                                      gis=gis,
                                      name=name,
                                      info=info,
                                      portaladmin=portaladmin)
        self._info = info
        initialize = kwargs.pop('initialize', False)
        self._name = name
        self._url = url
        self._pa = portaladmin
        self._con = portaladmin._con
        self._gis = gis
        if initialize:
            self._init()
    #----------------------------------------------------------------------
    def _init(self, connection=None):
        self._properties = PropertyMap(self._info)
        self._json_dict = self._info
    #----------------------------------------------------------------------
    def status(self):
        """
        This operation checks whether a portal machine is ready to receive
        requests.
        """
        url = "%s/status/%s" % (self._url, self._name)
        params = {"f": "json"}
        res = self._con.get(path=url,
                            params=params)
        if 'status' in res:
            return res['status'] == 'success'
        return False
    #----------------------------------------------------------------------
    def unregister(self):
        """
        This operation unregisters a portal machine from a portal site. The
        operation can only performed when there are two machines
        participating in a portal site.
        """
        params = {"f" : "json",
                  "machineName" : self._name}
        url = "%s/machines/unregister" % self._url
        return self._con.post(path=url, postdata=params)



try:
    import os, sys, json
    import numpy as np
    import torch
    import torch.nn as nn
    import math
    HAS_TORCH = True

    prf_root_dir = os.path.join(os.path.dirname(__file__), os.pardir)
    sys.path.append(prf_root_dir)

    sys.path.append(os.path.dirname(__file__))

    import util
except Exception as e:
    HAS_TORCH = False


def convert_bounding_boxes_to_coord_list(bounding_boxes):
    '''
    convert bounding box numpy array to python list of point arrays
    :param bounding_boxes: numpy array of shape [n, 4]
    :return: python array of point numpy arrays, each point array is in shape [4,2]
    '''
    num_bounding_boxes = bounding_boxes.shape[0]
    bounding_box_coord_list = []
    for i in range(num_bounding_boxes):
        coord_array = np.empty(shape=(4, 2), dtype=np.float)
        coord_array[0][0] = bounding_boxes[i][0]
        coord_array[0][1] = bounding_boxes[i][1]

        coord_array[1][0] = bounding_boxes[i][0]
        coord_array[1][1] = bounding_boxes[i][3]

        coord_array[2][0] = bounding_boxes[i][2]
        coord_array[2][1] = bounding_boxes[i][3]

        coord_array[3][0] = bounding_boxes[i][2]
        coord_array[3][1] = bounding_boxes[i][1]

        bounding_box_coord_list.append(coord_array)

    return bounding_box_coord_list


def calculate_rectangle_size_from_batch_size(batch_size):
    '''
    calculate number of rows and cols to composite a rectangle given a batch size
    :param batch_size:
    :return: number of cols and number of rows
    '''
    rectangle_height = int(math.sqrt(batch_size) + 0.5)
    rectangle_width = int(batch_size / rectangle_height)

    if rectangle_height * rectangle_width > batch_size:
        if rectangle_height >= rectangle_width:
            rectangle_height = rectangle_height - 1
        else:
            rectangle_width = rectangle_width - 1

    if (rectangle_height + 1) * rectangle_width <= batch_size:
        rectangle_height = rectangle_height + 1
    if (rectangle_width + 1) * rectangle_height <= batch_size:
        rectangle_width = rectangle_width + 1

    # swap col and row to make a horizontal rect
    if rectangle_height > rectangle_width:
        rectangle_height, rectangle_width = rectangle_width, rectangle_height

    if rectangle_height * rectangle_width != batch_size:
        return batch_size, 1

    return rectangle_height, rectangle_width


def get_tile_size(model_height, model_width, padding, batch_height, batch_width):
    '''
    Calculate request tile size given model and batch dimensions
    :param model_height:
    :param model_width:
    :param padding:
    :param batch_width:
    :param batch_height:
    :return: tile height and tile width
    '''
    tile_height = (model_height - 2 * padding) * batch_height
    tile_width = (model_width - 2 * padding) * batch_width

    return tile_height, tile_width


def tile_to_batch(pixel_block, model_height, model_width, padding, fixed_tile_size=True, **kwargs):
    inner_width = model_width - 2 * padding
    inner_height = model_height - 2 * padding

    band_count, pb_height, pb_width = pixel_block.shape
    pixel_type = pixel_block.dtype

    if fixed_tile_size is True:
        batch_height = kwargs['batch_height']
        batch_width = kwargs['batch_width']
    else:
        batch_height = math.ceil((pb_height - 2 * padding) / inner_height)
        batch_width = math.ceil((pb_width - 2 * padding) / inner_width)

    batch = np.zeros(shape=(batch_width * batch_height, band_count, model_height, model_width), dtype=pixel_type)
    for b in range(batch_width * batch_height):
        y = int(b / batch_width)
        x = int(b % batch_width)

        # pixel block might not be the shape (band_count, model_height, model_width)
        sub_pixel_block = pixel_block[:, y * inner_height: y * inner_height + model_height,
                    x * inner_width: x * inner_width + model_width]
        sub_pixel_block_shape = sub_pixel_block.shape
        batch[b, :, :sub_pixel_block_shape[1], :sub_pixel_block_shape[2]] = sub_pixel_block

    return batch, batch_height, batch_width


def batch_to_tile(batch, batch_height, batch_width):
    batch_size, bands, inner_height, inner_width = batch.shape
    tile = np.zeros(shape=(bands, inner_height * batch_height, inner_width * batch_width), dtype=batch.dtype)

    for b in range(batch_width * batch_height):
        y = int(b / batch_width)
        x = int(b % batch_width)

        tile[:, y * inner_height: (y+1) * inner_height, x * inner_width:(x+1) * inner_width] = batch[b]

    return tile


def remove_bounding_boxes_in_padding(bounding_boxes, scores, classes, image_height, image_width, padding,
                                     batch_height=1, batch_width=1):
    '''

    :param bounding_boxes: the batch of bounding boxes, shape=[B,N,4]
    :param scores: the batch of box scores, shape=[B,N]
    :param classes: the batch of labels, shape=[B,N]
    :param image_height: model height
    :param image_width: model width
    :param padding:
    :param batch_height:
    :param batch_width:
    :return:
    '''
    keep_indices = np.where((bounding_boxes[:,:,0] < image_height-padding) &
              (bounding_boxes[:,:,1] < image_width-padding) &
              (bounding_boxes[:,:,2] > padding) &
              (bounding_boxes[:,:,3] > padding))

    inner_width = image_width - 2 * padding
    inner_height = image_height - 2 * padding

    # convert coordinates in the batch to super tile and then filter by the keep_indices
    for b in range(batch_width * batch_height):
        y = int(b / batch_width)
        x = int(b % batch_width)

        bounding_boxes[b, :, [0, 2]] = bounding_boxes[b, :, [0, 2]] + y * inner_height
        bounding_boxes[b, :, [1, 3]] = bounding_boxes[b, :, [1, 3]] + x * inner_width

    bounding_boxes = bounding_boxes[keep_indices]
    scores = scores[keep_indices]
    classes = classes[keep_indices]

    return bounding_boxes, scores, classes

class ChildObjectDetector:

    def initialize(self, model, model_as_file):

        if not HAS_TORCH:
            raise Exception('PyTorch is not installed. Install it using conda install -c pytorch pytorch torchvision')

        from arcgis.learn.models import SingleShotDetector

        if torch.cuda.is_available():
            self.device = torch.device('cuda')
        else:
            self.device = torch.device('cpu')

        if model_as_file:
            with open(model, 'r') as f:
                self.json_info = json.load(f)
        else:
            self.json_info = json.load(model)

        model_path = self.json_info['ModelFile']
        if model_as_file and not os.path.isabs(model_path):
            model_path = os.path.abspath(os.path.join(os.path.dirname(model), model_path))

        self.ssd = SingleShotDetector.from_emd(data=None, emd_path=model)
        self.model = self.ssd.learn.model.to(self.device)
        self.model.eval()

        
    def getParameterInfo(self, required_parameters):
        required_parameters.extend(
            [
                {
                    'name': 'padding',
                    'dataType': 'numeric',
                    'value': 0,
                    'required': False,
                    'displayName': 'Padding',
                    'description': 'Padding'
                },
                {
                    'name': 'threshold',
                    'dataType': 'numeric',
                    'value': 0.5,
                    'required': False,
                    'displayName': 'Confidence Score Threshold [0.0, 1.0]',
                    'description': 'Confidence score threshold value [0.0, 1.0]'
                },
                {
                    'name': 'nms_overlap',
                    'dataType': 'numeric',
                    'value': 0.1,
                    'required': False,
                    'displayName': 'NMS Overlap',
                    'description': 'Maximum allowed overlap within each chip'
                },
                {
                    'name': 'batch_size',
                    'dataType': 'numeric',
                    'required': False,
                    'value': 64,
                    'displayName': 'Batch Size',
                    'description': 'Batch Size'
                }
            ]
        )
        return required_parameters

    def getConfiguration(self, **scalars):
        self.padding = int(scalars['padding'])
        self.nms_overlap = float(scalars['nms_overlap'])
        self.thres = float(scalars['threshold'])
        self.batch_size = int(scalars['batch_size'])

        self.rectangle_height, self.rectangle_width = calculate_rectangle_size_from_batch_size(self.batch_size)
        ty, tx = get_tile_size(self.json_info['ImageHeight'], self.json_info['ImageWidth'],
                                         self.padding, self.rectangle_height, self.rectangle_width)

        return {
            'extractBands': tuple(self.json_info['ExtractBands']),
            'padding': int(scalars['padding']),
            'threshold': self.thres,
            'nms_overlap': self.nms_overlap,
            'tx': tx,
            'ty': ty,
            'fixedTileSize': 1
        }

    def vectorize(self, **pixelBlocks): # 8 x 224 x 224 x 3
        input_image = pixelBlocks['raster_pixels']
        batch, batch_height, batch_width = \
            tile_to_batch(input_image,
                                    self.json_info['ImageHeight'],
                                    self.json_info['ImageWidth'],
                                    self.padding,
                                    fixed_tile_size=True,
                                    batch_height=self.rectangle_height,
                                    batch_width=self.rectangle_width)

        bounding_boxes, scores, classes = util.detect_objects_image_space(self.model, batch, self.ssd._anchors, self.ssd._grid_sizes, self.device,\
                                                                        classes=[clas['Name'] for clas in self.json_info['Classes']],\
                                                                        nms_overlap=self.nms_overlap, thres=self.thres)

        return convert_bounding_boxes_to_coord_list(bounding_boxes), scores, classes



""" Utility functions internally used by the store.
"""

from arcgis.features import FeatureSet

from ... import workforce


def add_features(feature_layer, features, use_global_ids=False):
    """ Adds features to the feature_layer.  The input features will be updated upon successful
        adding on the server, such that they contain the server-assigned object_ids and global_ids.
        :param feature_layer: An arcgis.features.FeatureLayer.
        :param features: list of arcgis.features.Features.
        :param use_global_ids: use global ids or not
        :returns: The added features.
        :raises ServerError: Indicates that the server rejected the new features.
    """
    if features:
        feature_set = FeatureSet(features)
        response = feature_layer.edit_features(adds=feature_set, use_global_ids=use_global_ids)
        add_results = response['addResults']
        errors = [result['error'] for result in add_results if not result['success']]
        if errors:
            raise workforce.ServerError(errors)
        for feature, add_results in zip(features, add_results):
            feature.attributes[feature_layer.properties('objectIdField')] = add_results['objectId']
            feature.attributes[feature_layer.properties('globalIdField')] = add_results['globalId']
    return features


def update_features(feature_layer, features):
    """ Updates features in a feature_layer.
        :param feature_layer: An arcgis.features.FeatureLayer.
        :param features: list of arcgis.features.Features.  Each feature must have an object id.
        :raises ServerError: Indicates that the server rejected the updates.
    """
    if features:
        response = feature_layer.edit_features(updates=FeatureSet(features))
        errors = [result['error'] for result in response["updateResults"] if not result['success']]
        if errors:
            raise workforce.ServerError(errors)
    return features


def remove_features(feature_layer, features):
    """ Removes features from a feature_layer.
        :param feature_layer: An arcgis.features.FeatureLayer.
        :param features: list of arcgis.features.Features.  Each feature must have an object id.
        :raises ServerError: Indicates that the server rejected the removals.
    """
    if features:
        object_id_attr = feature_layer.properties['objectIdField']
        object_ids = ','.join([str(feature.attributes[object_id_attr]) for feature in features])
        response = feature_layer.edit_features(deletes=object_ids)
        errors = [result['error'] for result in response['deleteResults'] if not result['success']]
        if errors:
            raise workforce.ServerError(errors)


def validate(validate_fn, **kwargs):
    """ Runs a Model validation routine, and raises the first ValidationError if any are returned.
        :param validate_fn: A function that takes no positional arguments, and returns a list of
        ValidationErrors.  Any **kwargs given to this function will be passed to validate_fn.
        :raises ValidationError: Indicates that the validation_fn returned a ValidationError.
    """
    validation_failures = validate_fn(**kwargs)
    if validation_failures:
        raise validation_failures[0]
